/**
 * Copyright (c) 2014-2016 Linagora
 * 
 * This program/library is free software: you can redistribute it and/or modify
 * it under the terms of the GNU Lesser General Public License as published by
 * the Free Software Foundation, either version 2.1 of the License, or (at your
 * option) any later version.
 * 
 * This program/library is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or
 * FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser General Public License
 * for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program/library; If not, see http://www.gnu.org/licenses/
 * for the GNU Lesser General Public License version 2.1.
 */
package org.ow2.petals.microkernel.api.util;

import java.net.URI;
import java.net.URISyntaxException;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.ow2.petals.clientserverapi.configuration.ContainerConfiguration;
import org.ow2.petals.microkernel.api.configuration.ConfigurationService;
import org.ow2.petals.microkernel.api.configuration.DomainConfiguration;
import org.ow2.petals.microkernel.api.configuration.JndiConfiguration;
import org.ow2.petals.microkernel.api.configuration.exception.ConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.ContainerNameDuplicatedException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerDuplicatedPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerFastRemoteTcpPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerHostException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerJmxPortException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerNameException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerPasswordException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerTransportSvcException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentContainerUserException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainConfigurationException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiFactoryException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainJndiProviderUrlException;
import org.ow2.petals.microkernel.api.configuration.exception.InconsistentDomainModeException;
import org.ow2.petals.topology.TopologyHelper;
import org.ow2.petals.topology.generated.Container;
import org.ow2.petals.topology.generated.Domain;
import org.ow2.petals.topology.generated.DomainMode;
import org.ow2.petals.topology.generated.JmxService;
import org.ow2.petals.topology.generated.Jndi;
import org.ow2.petals.topology.generated.Topology;
import org.ow2.petals.topology.generated.TransportService;

import com.ebmwebsourcing.easycommons.lang.StringHelper;
import com.ebmwebsourcing.easycommons.lang.UncheckedException;

/**
 * Utilities about {@link ConfigurationService}
 * 
 * @author Christophe DENEUX - Linagora
 * 
 */
public class ConfigurationServiceUtils {

    private ConfigurationServiceUtils() {
        // No constructor, it's an utility class
    }

    /**
     * Initializes a container configuration ({@link ContainerConfiguration}) bean from a container JAXB element (
     * {@link Container})
     * 
     * @param containerConfiguration
     *            The container configuration to initialize from the container <code>container</code>. Can not be
     *            {@code null}.
     * @param container
     *            A container configuration coming from its XML definition to use to initialize the container
     *            configuration <code>containerConfiguration</code>. Can not be <code>null</code>.
     * @param domainName
     *            The name of domain whose the container is a member
     */
    public static void initializeContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final Container container, final String domainName) {

        containerConfiguration.setName(container.getName());
        containerConfiguration.setDescription(container.getDescription());
        containerConfiguration.setUser(container.getUser());
        containerConfiguration.setPassword(container.getPassword());
        containerConfiguration.setHost(container.getHost());

        // JMX Service
        if (container.getJmxService() != null) {
            containerConfiguration.setJmxRMIConnectorPort(container.getJmxService().getRmiPort());
        }

        // Transport Service
        final TransportService transportService = container.getTransportService();
        if (transportService != null) {
            containerConfiguration.setTCPPort(transportService.getTcpPort());
        }

        containerConfiguration.setDomainName(domainName);
    }

    public static void initializeTopology(final Topology topology, final DomainConfiguration domainConfiguration,
            final Collection<ContainerConfiguration> containerConfigurations) {
        Domain domain = new Domain();
        ConfigurationServiceUtils.initializeDomain(domain, domainConfiguration);
        topology.setDomain(domain);
        for (final ContainerConfiguration containerConfiguration : containerConfigurations) {
            final Container container = new Container();
            ConfigurationServiceUtils.initializeContainer(container, containerConfiguration);
            domain.getContainer().add(container);
        }
    }

    public static void initializeDomain(final Domain domain, final DomainConfiguration domainConfiguration) {
        domain.setName(domainConfiguration.getName());
        domain.setDescription(domainConfiguration.getDescription());
        if (domainConfiguration.getMode() == DomainConfiguration.DomainMode.DYNAMIC) {
            domain.setMode(DomainMode.DYNAMIC);
        } else if (domainConfiguration.getMode() == DomainConfiguration.DomainMode.STATIC) {
            domain.setMode(DomainMode.STATIC);
        } else {
            throw new UncheckedException("Impossible case: unknown domain mode");
        }
        domain.setRegistryImplementation(domainConfiguration.getSharedAreaImplementation());
        final List<Object> extraParameters = domainConfiguration.getExtraParameters();
        if (extraParameters != null) {
            domain.getAny().addAll(extraParameters);
        }
        JndiConfiguration jndiConfiguration = domainConfiguration.getJndiConfiguration();
        if (jndiConfiguration != null) {
            final Jndi jndi = new Jndi();
            domain.setJndi(jndi);
            jndi.setBatchSize(jndiConfiguration.getJndiBatchSize());
            jndi.setFactory(jndiConfiguration.getJndiFactory());
            jndi.setPoolSize(jndiConfiguration.getJndiPoolSize());
            jndi.setProviderUrl(jndiConfiguration.getJndiProviderUrl().toString());
            jndi.setSecurityCredentials(jndiConfiguration.getJndiSecurityCredentials());
            jndi.setSecurityPrincipal(jndiConfiguration.getJndiSecurityPrincipal());
        }

    }

    public static void initializeContainer(final Container container,
            final ContainerConfiguration containerConfiguration) {

        container.setName(containerConfiguration.getName());
        container.setDescription(containerConfiguration.getDescription());
        container.setHost(containerConfiguration.getHost());
        container.setUser(containerConfiguration.getUser());
        container.setPassword(containerConfiguration.getPassword());

        final JmxService jmxService = new JmxService();
        jmxService.setRmiPort(containerConfiguration.getJmxRMIConnectorPort());

        container.setJmxService(jmxService);

        final TransportService transportService = new TransportService();
        transportService.setTcpPort(containerConfiguration.getTCPPort());

        container.setTransportService(transportService);
    }

    /**
     * Initializes a domain configuration from a topology JAXB element.
     * 
     * @param domainConfiguration
     *            The domain configuration to initialize from the topology <code>topology</code>. Can not be
     *            {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition to use to initialize the domain configuration
     *            <code>domainConfiguration</code>. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is missing
     */
    public static void initializeDomainConfiguration(final DomainConfiguration domainConfiguration,
            final Topology topology) throws ConfigurationException {
        // domain mode
        final Domain domain = topology.getDomain();
        domainConfiguration.setName(domain.getName());
        domainConfiguration.setDescription(domain.getDescription());

        if (DomainMode.STATIC.equals(domain.getMode())) {
            domainConfiguration.setMode(DomainConfiguration.DomainMode.STATIC);
        } else if (DomainMode.DYNAMIC.equals(domain.getMode())) {
            domainConfiguration.setMode(DomainConfiguration.DomainMode.DYNAMIC);
        } else {
            throw new UncheckedException("Impossible case: unknown domain mode");
        }
        if (domain.getRegistryImplementation() != null) {
            domainConfiguration.setSharedAreaImplementation(domain.getRegistryImplementation().trim());
        }
        if (domain.getAny() != null) {
            domainConfiguration.setExtraParameters(domain.getAny());
        }

        // domain JNDI
        if (domain.getJndi() != null) {
            final JndiConfiguration jndiConfiguration = new JndiConfiguration();

            jndiConfiguration.setJndiFactory(domain.getJndi().getFactory());
            try {
                jndiConfiguration.setJndiProviderUrl(new URI(domain.getJndi().getProviderUrl()));
            } catch (URISyntaxException e) {
                throw new ConfigurationException(e);
            }
            jndiConfiguration.setJndiSecurityPrincipal(domain.getJndi().getSecurityPrincipal());
            jndiConfiguration.setJndiSecurityCredentials(domain.getJndi().getSecurityCredentials());

            final Integer poolSize = domain.getJndi().getPoolSize();
            if (poolSize != null) {
                jndiConfiguration.setJndiPoolSize(poolSize.intValue());
            }
            final Integer batchSize = domain.getJndi().getBatchSize();
            if (batchSize != null) {
                jndiConfiguration.setJndiBatchSize(batchSize.intValue());
            }

            domainConfiguration.setJndiConfiguration(jndiConfiguration);
        }
    }

    /**
     * Validate the consistency of the topology against a container configuration.
     * 
     * @param containerConfiguration
     *            A container configuration to validate against the topology <code>topology</code>. Can not be
     *            {@code null}
     * @param topology
     *            The topology to validate against the container configuration <code>containerConfiguration</code>. Can
     *            not be <code>null</code>.
     * @throws InconsistentConfigurationException
     *             The configuration is inconsistent.
     */
    public static void validateTopology(final ContainerConfiguration containerConfiguration, final Topology topology)
            throws InconsistentConfigurationException {

        // Checks if the current container name can be found in the topology
        final String containerName = containerConfiguration.getName();
        if (TopologyHelper.findContainer(containerName, topology) == null) {
            throw new InconsistentContainerNameException(containerName);
        }
        // the container name has been found in the topology

        // validate JNDI consistency
        final Domain domain = topology.getDomain();
        final Jndi jndi = domain.getJndi();
        if (jndi != null) {
            final String jndiFactory = jndi.getFactory();
            if (jndiFactory == null || jndiFactory.trim().isEmpty()) {
                throw new InconsistentDomainJndiFactoryException();
            }
            if (jndi.getProviderUrl() == null) {
                throw new InconsistentDomainJndiProviderUrlException();
            }
        }

        // Validate the domain mode
        final DomainMode domainMode = domain.getMode();
        if (!DomainMode.STATIC.equals(domainMode) && !DomainMode.DYNAMIC.equals(domainMode)) {
            throw new InconsistentDomainModeException(domain.getName());
        }

        // validate containers
        validateTopologyContainers(domain.getContainer());

    }

    /**
     * Validate the consistency of the topology containers
     * 
     * @param containers
     *            The topology containers to validate. Can not be <code>null</code>.
     * @throws InconsistentConfigurationException
     *             The configuration is inconsistent.
     */
    public static void validateTopologyContainers(final List<Container> containers)
            throws InconsistentConfigurationException {

        final Map<String, Set<Integer>> networkPortAlreaySet = new HashMap<>();

        final Collection<String> containerNames = new HashSet<String>();
        for (final Container container : containers) {

            if (!containerNames.add(container.getName())) {
                throw new ContainerNameDuplicatedException(container.getName());
            }

            final String containerHost = container.getHost();
            if (containerHost == null) {
                throw new InconsistentContainerHostException(container.getName());
            }

            if (containers.size() > 1 && container.getTransportService() == null) {
                throw new InconsistentContainerTransportSvcException(container.getName());
            }

            // Validate network ports
            final Set<Integer> containerNetworkPortAlreaySet;
            if (!networkPortAlreaySet.containsKey(containerHost)) {
                containerNetworkPortAlreaySet = new HashSet<Integer>();
                networkPortAlreaySet.put(containerHost, containerNetworkPortAlreaySet);
            } else {
                containerNetworkPortAlreaySet = networkPortAlreaySet.get(containerHost);
            }

            if (container.getJmxService().getRmiPort() != 0) {
                if (!containerNetworkPortAlreaySet.add(container.getJmxService().getRmiPort())) {
                    throw new InconsistentContainerDuplicatedPortException(container.getName(),
                            container.getJmxService().getRmiPort());
                }
            }

            if ((container.getTransportService() != null) && (container.getTransportService().getTcpPort() >= 0)) {
                if (!containerNetworkPortAlreaySet.add(container.getTransportService().getTcpPort())) {
                    throw new InconsistentContainerDuplicatedPortException(container.getName(),
                            container.getTransportService().getTcpPort());
                }
            }

        }
    }

    public static void verifyContainer(final ContainerConfiguration newContainerConfiguration,
            ContainerConfiguration containerConfig) throws InconsistentContainerHostException,
            InconsistentContainerUserException, InconsistentContainerPasswordException,
            InconsistentContainerJmxPortException, InconsistentContainerFastRemoteTcpPortException {
        // TODO is that the correct way to compare hosts...
        if (!StringHelper.equal(containerConfig.getHost(), newContainerConfiguration.getHost())) {
            throw new InconsistentContainerHostException(newContainerConfiguration.getName());
        }

        if (!StringHelper.equal(containerConfig.getUser(), newContainerConfiguration.getUser())) {
            throw new InconsistentContainerUserException(newContainerConfiguration.getName());
        }
        if (!StringHelper.equal(containerConfig.getPassword(), newContainerConfiguration.getPassword())) {
            throw new InconsistentContainerPasswordException(newContainerConfiguration.getName());
        }

        // JMX Service
        if (containerConfig.getJmxRMIConnectorPort() != newContainerConfiguration.getJmxRMIConnectorPort()) {
            throw new InconsistentContainerJmxPortException(newContainerConfiguration.getName());
        }

        // Transport service
        if (containerConfig.getTCPPort() != newContainerConfiguration.getTCPPort()) {
            throw new InconsistentContainerFastRemoteTcpPortException(newContainerConfiguration.getName());
        }
    }

    /**
     * Verify the given new domain configuration against another configuration.
     * 
     * @param newDomainConfiguration
     *            The new domain configuration
     * @throws InconsistentDomainConfigurationException
     *             Something is incompatible between the existing domain configuration and the new one
     */
    public static void verifyDomainConfiguration(final DomainConfiguration newDomainConfiguration,
            final DomainConfiguration oldDomainConfiguration) throws InconsistentDomainConfigurationException {

        if (!newDomainConfiguration.getMode().equals(oldDomainConfiguration.getMode())) {
            throw new InconsistentDomainModeException(newDomainConfiguration.getName());
        }

        // JNDI service
        final JndiConfiguration jndiConfiguration = oldDomainConfiguration.getJndiConfiguration();
        final JndiConfiguration newJndiConfiguration = newDomainConfiguration.getJndiConfiguration();
        if ((jndiConfiguration == null && newJndiConfiguration != null)
                || (jndiConfiguration != null && newJndiConfiguration == null)) {
            throw new InconsistentDomainConfigurationException("JNDI configuration mismatch");
        } else if (jndiConfiguration != null && newJndiConfiguration != null) {
            if (!StringHelper.equal(newJndiConfiguration.getJndiFactory(), jndiConfiguration.getJndiFactory())) {
                throw new InconsistentDomainJndiFactoryException();
            }
            if ((newJndiConfiguration.getJndiProviderUrl() == null && jndiConfiguration.getJndiProviderUrl() != null)
                    || (newJndiConfiguration.getJndiProviderUrl() != null)
                            && (jndiConfiguration.getJndiProviderUrl() == null)
                    || (newJndiConfiguration.getJndiProviderUrl() != null
                            && jndiConfiguration.getJndiProviderUrl() != null && !newJndiConfiguration
                                    .getJndiProviderUrl().equals(jndiConfiguration.getJndiProviderUrl()))) {
                throw new InconsistentDomainJndiProviderUrlException();
            }
            if (!StringHelper.equal(newJndiConfiguration.getJndiSecurityPrincipal(),
                    jndiConfiguration.getJndiSecurityPrincipal())) {
                throw new InconsistentDomainConfigurationException("JNDI Security Principal mismatch");
            }
            if (!StringHelper.equal(newJndiConfiguration.getJndiSecurityCredentials(),
                    jndiConfiguration.getJndiSecurityCredentials())) {
                throw new InconsistentDomainConfigurationException("JNDI Security Credentials mismatch");
            }
            if (newJndiConfiguration.getJndiPoolSize() != jndiConfiguration.getJndiPoolSize()) {
                throw new InconsistentDomainConfigurationException("JNDI Pool Size mismatch");
            }
            if (newJndiConfiguration.getJndiBatchSize() != jndiConfiguration.getJndiBatchSize()) {
                throw new InconsistentDomainConfigurationException("JNDI Batch Size mismatch");
            }
        }
    }

    /**
     * <p>
     * Initializes a container configuration bean from a topology JAXB element.
     * </p>
     * <p>
     * To be able to retrieve the container into the topology, its name <b>MUST<b> initialized in
     * <code>containerConfiguration</code>.
     * </p>
     * 
     * @param containerConfiguration
     *            The container configuration to initialize from the topology <code>topology</code>. Can not be
     *            {@code null}.
     * @param topology
     *            A topology configuration coming from its XML definition to use to initialize the container
     *            configuration <code>containerConfiguration</code>. Can not be <code>null</code>.
     * @throws ConfigurationException
     *             Something is missing
     */
    public static void initializeContainerConfiguration(final ContainerConfiguration containerConfiguration,
            final Topology topology) throws ConfigurationException {

        final String containerName = containerConfiguration.getName();
        final Container localContainer = TopologyHelper.findContainer(containerName, topology);
        if (localContainer == null) {
            throw new InconsistentContainerNameException(containerName);
        }

        initializeContainerConfiguration(containerConfiguration, localContainer, topology.getDomain().getName());
    }
}
